/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <stdio.h>

#include "gtkeditor.h"
#include "internal.h"
#include "scanner.h"
#include "highlight.h"
#include "edit.h"

#define TOKEN_LOOKBACK 5
#define BEGIN_BLOCK 0
#define END_BLOCK 1

typedef struct _TokenMatch {
  GList *token;
  int pos;
  GNode *cur_block;
  GNode *prev_block;
  GNode *next_block;
}TokenMatch;

/*
 * globals
 */

/* pointer to current editor being analized */
static GtkEditor *g_editor;

/* text changes associated with the newly inserted text */
static GSList *g_new_text_changes = NULL;

/* block stuff */
static Block *g_block, *g_next;
static GList *g_block_token;
static int g_block_token_pos;
static GNode *g_base_block, *g_base_prev, *g_base_next;;

/* used for recyclying */
static GList *g_free_tokens = NULL;
static GNode *g_free_blocks = NULL;

/*
 * prototypes
 */
static void          delete_blocks         (GNode *block);
static TokenMatch    *find_nearest_token   (GtkEditor *editor,
                                            int pos);
static void           look_back            (TokenMatch *token_match,
                                            int count);
static void           split_token          (GList          *token,
					    int             pos);
static GList         *get_new_token        (void);


#if DEBUG
static void print_tokens (GtkEditor *editor);
#endif

/* gtk_editor_insert_text

   This function is called when new text is inserted by the user. It
   scans the new text with the effected areas, and updates the editors
   token list.
*/
void
gtk_editor_insert_text (GtkEditable *editable, const gchar *new_text,
			gint new_text_length, gint *position)
{
  GtkEditor *editor = GTK_EDITOR (editable);
  GtkText *text = GTK_TEXT (editable);
  GNode *temp_block;
  Token *token;
  Block *block;
  TokenMatch *token_match = NULL;
  int pos;

  if (!editor->scanner) {
    /* No scanner installed insert with the default font properties */
    gtk_text_set_point (text, *position);
    gtk_text_insert (text, GTK_WIDGET (editable)->style->font,
		     &GTK_WIDGET (editable)->style->text[GTK_STATE_NORMAL],
		     &GTK_WIDGET (editable)->style->base[GTK_STATE_NORMAL],
		     new_text, new_text_length);
    *position += new_text_length;
    gtk_signal_emit_stop_by_name (GTK_OBJECT(editable), "insert_text");
    return;
  }

  if (!editor->tokens && gtk_text_get_length (text)) {
    /* Uninitialized editor, highlight the entire buffer before inserting
       anything.
    */
    gtk_editor_hilite_buffer (editor);
  }

  /*
     Create a new token and insert it into the token list. This token
     represents the inserted text. A prop_id of -1 represents a token
     with no associated text in the editor. This fact is used in the
     merge routine below.
  */
  g_token = get_new_token ();
  token = g_token->data;
  token->id = -1;
  token->prop_id = -1;
  token->length = new_text_length;

  g_block_token = NULL;
  if (!editor->tokens) {
    /* empty editor */
    editor->tokens = g_token;
    g_token->next = NULL;
    g_token->prev = NULL;
  } else {
    /* find insertion point */
    token_match = find_nearest_token(editor, *position);

    /* insert the token into the proper position */
    if (*position == 0) {
      g_token->next = editor->tokens;
      editor->tokens = g_token;
      g_token->prev = NULL;
      g_token->next->prev = g_token;
      token_match->next_block = token_match->cur_block;
      token_match->cur_block = NULL;
      token_match->prev_block = NULL;
    } else if (*position == gtk_text_get_length(GTK_TEXT(editor))) {
      g_token->prev = token_match->token;
      g_token->next = NULL;
      token_match->token->next = g_token;
    } else {
      if (token_match->pos != *position) {
        if (token_match->cur_block) {
          block = token_match->cur_block->data;
          if (token_match->token == block->begin) {
            /* token needs to be split, destroy the block */
            temp_block = token_match->cur_block;
            token_match->prev_block = token_match->cur_block->prev;
            token_match->next_block = token_match->cur_block->next;
            token_match->cur_block = token_match->cur_block->parent;
            _gtk_editor_remove_block (editor, temp_block);
            g_base_block = token_match->cur_block;
            g_base_prev = token_match->prev_block;
            g_base_next = token_match->next_block;
            g_block_token = token_match->token;
            g_block_token_pos = token_match->pos;
          } else if (token_match->token == block->end) {
            block->end = NULL;
            g_base_block = token_match->cur_block;
            g_base_prev = token_match->prev_block;
            g_base_next = token_match->next_block;
            g_block_token = token_match->token;
            g_block_token_pos = token_match->pos;
          }
        }
        split_token(token_match->token, *position - token_match->pos);
        token_match->token = token_match->token->next;
      }
      g_token->next = token_match->token;
      g_token->prev = token_match->token->prev;
      if(token_match->token->prev)
        token_match->token->prev->next = g_token;
      token_match->token->prev = g_token;
    }
  }

  /* look back a certain ammount */
  if(token_match) {
    token_match->token = g_token;
    token_match->pos = *position;
    look_back(token_match, TOKEN_LOOKBACK);
    g_token = token_match->token;
    pos = token_match->pos;
    g_cur_block = token_match->cur_block;
    g_next_block = token_match->next_block;
    g_prev_block = token_match->prev_block;
  } else {
    pos = *position;
    g_cur_block = NULL;
    g_next_block = NULL;
    g_prev_block = NULL;
  }

  /* scan the text */
  g_new_text_changes = NULL;
  g_editor = editor;
  if (g_cur_block)
    g_block = g_cur_block->data;
  if (g_next_block)
    g_next = g_next_block->data;

  _gtk_editor_scan (editor, pos, new_text, new_text_length, *position);
  /* Insert the new text with the new text properties. The reason I do this
     after I scan the text is because of a bug in the text widget.
  */
  _gtk_editor_insert_text_with_properties (editor, new_text,
					   g_new_text_changes);
    /* apply block changes */
  if(g_block_token)
    _gtk_editor_update_blocks (editor, g_base_block,
			       g_base_prev, g_base_next,
			       g_block_token, g_block_token_pos);
  /* set the new position*/
  *position += new_text_length;
  gtk_text_set_point (GTK_TEXT(editor), *position);
  gtk_editable_set_position (editable, *position);
  /* terminate this signal */
  gtk_signal_emit_stop_by_name (GTK_OBJECT(editable), "insert_text");
}

/* gtk_editor_delete_text

   This function is called when the user deletes text. It deletes the
   asociated tokens and block and rescans the affected area.

*/
void
gtk_editor_delete_text (GtkEditable *editable,
			gint start_pos, gint end_pos)
{
  GtkEditor *editor = GTK_EDITOR (editable);
  GList *prev = NULL, *temp, *cur = editor->cur_token;
  GNode *temp_block, *cur_block, *prev_block = NULL, *next_block;
  Token *token = cur->data;
  Block *block = NULL;
  TokenMatch *token_match;
  int pos = editor->cur_token_pos, prev_pos = 0;
  int deletion_length, flag = 0;

  /* calculate the length of the deleted area*/
  deletion_length = end_pos - start_pos;

  if (!editor->scanner) {
    /* No scanner just delete the text and return. */
    gtk_text_set_point(GTK_TEXT(editor), start_pos);
    gtk_text_forward_delete(GTK_TEXT(editor), deletion_length);
    gtk_signal_emit_stop_by_name (GTK_OBJECT(editable), "delete_text");
    return;
  }

  if((gtk_text_get_length(GTK_TEXT(g_editor)) - deletion_length) == 0) {
    /* cleared editor, get rid of everything */
    for(cur = editor->tokens; cur != NULL;) {
      temp = cur;
      cur = cur->next;
      temp->next = NULL;
      temp->prev = NULL;
      g_free_tokens = g_list_concat (temp, g_free_tokens);
    }
    editor->tokens = NULL;
    editor->cur_token = NULL;
    editor->cur_token_pos = 0;
    delete_blocks(editor->blocks);
    editor->blocks = NULL;
    editor->cur_block = NULL;
    editor->prev_block = NULL;
    editor->next_block = NULL;
    gtk_text_set_point (GTK_TEXT (editor), start_pos);
    gtk_text_forward_delete (GTK_TEXT (editor), deletion_length);
    gtk_signal_emit_stop_by_name (GTK_OBJECT(editable), "delete_text");
    return;
  }

  /* find nearest token */
  token_match = find_nearest_token(editor, start_pos);
  cur = token_match->token;
  token = cur->data;
  g_block_token = NULL;

  if (start_pos > token_match->pos) {
    /* starts in the middle of a token */
    if (token_match->cur_block) {
      block = token_match->cur_block->data;
      if (cur == block->begin) {
        /* token needs to be split, destroy the block */
        temp_block = token_match->cur_block;
        token_match->cur_block = token_match->cur_block->parent;
        _gtk_editor_remove_block(editor, temp_block);
        if(editor->blocks == NULL) {
          editor->cur_block = NULL;
          editor->prev_block = NULL;
          editor->next_block = NULL;
          token_match->cur_block = NULL;
          token_match->prev_block = NULL;
          token_match->next_block = NULL;
        }
        g_base_block = token_match->cur_block;
        g_base_prev = token_match->prev_block;
        g_base_next = token_match->next_block;
        g_block_token = token_match->token;
        g_block_token_pos = token_match->pos;
        flag = 1;
      } else if (cur == block->end) {
        block->end = NULL;
        g_base_block = token_match->cur_block;
        g_base_prev = token_match->prev_block;
        g_base_next = token_match->next_block;
        g_block_token = token_match->token;
        g_block_token_pos = token_match->pos;
        flag = 1;
      }
    }
    /* split the token */
    split_token (cur, start_pos - token_match->pos);
    cur = cur->next;
    token = cur->data;
  }

  /* set prev to the token prior to the deleted area */
  if (cur->prev) {
    prev = cur->prev;
    token = prev->data;
    prev_pos = start_pos - token->length;
  }

  cur_block = token_match->cur_block;
  if(cur_block)
    block = cur_block->data;
  next_block = token_match->next_block;
  token = cur->data;
  /* get rid of tokens and blocks in the deleted area */
  pos = start_pos;
  while (pos < end_pos) {
    if (cur_block && cur == block->begin) {
      temp_block = cur_block;
      if (token_match->cur_block == cur_block)
        token_match->cur_block = cur_block->parent;
      cur_block = cur_block->parent;
      if (cur_block)
        block = cur_block->data;
      if (!g_block_token) {
        g_base_block = cur_block;
        g_base_prev = prev_block;
        g_base_next = next_block;
        g_block_token = prev;
        g_block_token_pos = prev_pos;
        flag = 1;
      }
      _gtk_editor_remove_block(editor, temp_block);
    } else if (cur_block && cur == block->end) {
      block->end = NULL;
      if (!g_block_token) {
        g_base_block = token_match->cur_block;
        g_base_prev = token_match->prev_block;
        g_base_next = token_match->next_block;
        g_block_token = prev;
        g_block_token_pos = prev_pos;
        flag = 1;
      }
    }

    if (editor->blocks == NULL) {
      editor->cur_block = NULL;
      editor->prev_block = NULL;
      editor->next_block = NULL;
      cur_block = NULL;
      prev_block = NULL;
      next_block = NULL;
    }

    if (pos + token->length <= end_pos) {
      /* Token is entirely inside the deleted area, recycle it. */
      pos += token->length;
      temp = cur;
      if (cur->prev)
	cur->prev->next = cur->next;
      if (cur->next) {
	cur->next->prev = cur->prev;
	cur = cur->next;
	token = cur->data;
      }
      temp->next = NULL;
      temp->prev = NULL;
      g_free_tokens = g_list_concat (temp, g_free_tokens);
    } else {
      /* Token is covers more than the deleted are, shorten it. */
      token->length -= end_pos - pos;
      pos = end_pos;
    }
  }

  if (flag && prev == NULL) {
    g_block_token = cur;
    g_block_token_pos = end_pos;
  }

  pos = start_pos;
  if(prev) {
    cur = prev;
    token = cur->data;
    pos -= token->length;
  }

  /* look back a certain ammount */
  token_match->token = cur;
  token_match->pos = pos;
  look_back(token_match, TOKEN_LOOKBACK);

  /* rescan affected text */
  g_editor = editor;
  g_token = token_match->token;
  g_cur_block = token_match->cur_block;
  g_next_block = token_match->next_block;
  g_prev_block = token_match->prev_block;
  if (g_cur_block)
    g_block = g_cur_block->data;
  if (g_next_block)
    g_next = g_next_block->data;
  _gtk_editor_scan (editor, token_match->pos, NULL,
		    -deletion_length, start_pos);
  /* delete the deleted text */
  gtk_text_set_point (GTK_TEXT (editor), start_pos);
  gtk_text_forward_delete (GTK_TEXT (editor), deletion_length);
  /* apply block changes */
  if(g_block_token)
    _gtk_editor_update_blocks (editor, g_base_block,
			       g_base_prev, g_base_next,
			       g_block_token, g_block_token_pos);
  /* set the cursor position */
  gtk_text_set_point(GTK_TEXT(editor), start_pos);
  gtk_editable_set_position(editable, start_pos);
  /* terminate this signal */
  gtk_signal_emit_stop_by_name (GTK_OBJECT(editable), "delete_text");
}

/* gtk_editor_hilite_buffer

   Rescans the entire buffer and highlights acordingly.
*/
void
gtk_editor_hilite_buffer (GtkEditor *editor)
{
  int point, length;
  GtkEditable *editable;
  GtkText *text;
  char *buffer;
  Token *token;

  if (editor->scanner == NULL)
    /* nothing to do here */
    return;

  editable = GTK_EDITABLE (editor);
  text = GTK_TEXT (editor);

  if (!(length = gtk_text_get_length (text)))
    return;
  if (editor->tokens)
    g_list_free(editor->tokens);

  /*
     Create a new token an insert it into the token list. This token
     represents the entire buffer. A prop_id of -1 represents a token
     with no associated text in the editor. This fact is used in the
     merge routine below.
  */
  g_token = editor->tokens = get_new_token();
  token = g_token->data;
  token->id = -1;
  token->prop_id = -1;
  token->length = length;
  g_token->next = NULL;
  g_token->prev = NULL;

  gtk_text_freeze (text);

  buffer = gtk_editable_get_chars (editable, 0, length);
  gtk_text_set_point (text, 0);
  gtk_text_forward_delete (text, length);
  point = gtk_text_get_point (text);
  g_new_text_changes = NULL;
  g_editor = editor;
  g_cur_block = NULL;
  g_next_block = NULL;
  g_prev_block = NULL;
  g_block = NULL;
  g_next = NULL;
  _gtk_editor_scan (editor, 0, buffer, length, 0);
  _gtk_editor_insert_text_with_properties (editor, buffer,
					   g_new_text_changes);
  gtk_text_set_point (text, 0);
  gtk_editable_set_position (editable, 0);
  g_free(buffer);

  gtk_text_thaw (text);
}

static void
delete_blocks (GNode *block)
{
  GNode *cur, *temp;
  for(cur = block; cur != NULL;) {
    if(cur->children)
      delete_blocks(cur->children);
    temp = cur;
    cur = cur->next;
    temp->next = g_free_blocks;
    temp->parent = NULL;
    temp->children = NULL;
    temp->prev = NULL;
    g_free_blocks = temp;
  }
}

void
_gtk_editor_remove_block (GtkEditor *editor, GNode *block)
{
  GNode *temp;

  if (!block)
    return;

  if (block->children) {
    temp = block->children;
    temp->prev = block->prev;
    if (block->prev)
      block->prev->next = temp;
    temp->parent = block->parent;
    while(temp->next != NULL) {
      temp = temp->next;
      temp->parent = block->parent;
    }
    temp->parent = block->parent;
    temp->next = block->next;
    if(block->next)
      block->next->prev = temp;
  } else {
    if (block->prev)
      block->prev->next = block->next;
    if (block->next)
      block->next->prev = block->prev;
  }

  if (!block->parent && !block->prev)
    editor->blocks = block->children;


  block->next = g_free_blocks;
  block->parent = NULL;
  block->children = NULL;
  block->prev = NULL;
  g_free_blocks = block;
}

/* find_nearest_token

   Finds the nearest token and its coresponding blocks

*/
static TokenMatch *
find_nearest_token(GtkEditor *editor, int pos)
{
  GList *cur_token = editor->cur_token;
  GNode *cur_block = editor->cur_block,
        *prev_block = editor->prev_block,
        *next_block = editor->next_block;
  Token *token;
  Block *block = NULL, *prev = NULL, *next = NULL;
  int cur_pos = editor->cur_token_pos;
  static TokenMatch token_match;

  if (!cur_token) {
    /* empty editor */
    token_match.token = NULL;
    token_match.pos = 0;
    token_match.cur_block = NULL;
    token_match.prev_block = NULL;
    token_match.next_block = NULL;
    return &token_match;
  }

  /* initialize variables */
  if (cur_block)
    block = cur_block->data;
  token = cur_token->data;

  if (pos > cur_pos + token->length) {
    if (next_block)
      next = next_block->data;
    /* Position is greater than the current position, search forward. */
    while (cur_pos + token->length < pos) {
      /* check for crossing block boundries */
      /* increment the token */
      cur_pos += token->length;
      cur_token = cur_token->next;
      token = cur_token->data;
      if (cur_block && cur_token->prev == block->end) {
        next_block = cur_block->next;
        prev_block = cur_block;
        cur_block = cur_block->parent;
        if (cur_block)
          block = cur_block->data;
        if (next_block)
          next = next_block->data;
      }
      if (next_block && cur_token == next->begin) {
        cur_block = next_block;
        block = next;
        prev_block = NULL;
        next_block = cur_block->children;
        if (next_block)
          next = next_block->data;
      }
    }
  } else if(pos < cur_pos) {
    if (prev_block)
      prev = prev_block->data;
    /* Position is less than the current position, search back */
    while (cur_pos > pos) {
      /* decrement the token */
      cur_token = cur_token->prev;
      token = cur_token->data;
      cur_pos -= token->length;
      /* check for crossing block boundries */
      if (cur_block && cur_token->next == block->begin) {
        next_block = cur_block;
        prev_block = cur_block->prev;
        cur_block = cur_block->parent;
        if(prev_block)
          prev = prev_block->data;
        if (cur_block)
          block = cur_block->data;
      }
      if (prev_block && cur_token == prev->end) {
        cur_block = prev_block;
        block = prev;
        if (prev_block->children) {
          prev_block = g_node_last_sibling(prev_block->children);
          prev = prev_block->data;
        } else {
          prev_block = NULL;
        }
        next_block = NULL;
      }
    }
  }

  /* return the match */
  token_match.token = cur_token;
  token_match.pos = cur_pos;
  token_match.cur_block = cur_block;
  token_match.prev_block = prev_block;
  token_match.next_block = next_block;

  return &token_match;
}

static void
look_back (TokenMatch *token_match, int count)
{
  GList *cur_token = token_match->token;
  GNode *cur_block = token_match->cur_block,
        *prev_block = token_match->prev_block,
        *next_block = token_match->next_block;
  Token *token;
  Block *block = NULL, *prev = NULL;
  int pos = token_match->pos;

  /* initialize variables */
  if (cur_block)
    block = cur_block->data;
  if (prev_block)
    prev = prev_block->data;
  token = cur_token->data;

  /* look back */
  while (cur_token->prev != NULL && count > 0) {
    count--;
    /* check for crossing block boundries */
    if (cur_block && cur_token->next == block->begin) {
      next_block = cur_block;
      prev_block = cur_block->prev;
      cur_block = cur_block->parent;
      if(prev_block)
        prev = prev_block->data;
      if (cur_block)
        block = cur_block->data;
    }
    if (prev_block && cur_token == prev->end) {
      cur_block = prev_block;
      block = prev;
      if (prev_block->children) {
        prev_block = g_node_last_sibling(prev_block->children);
        prev = prev_block->data;
      } else {
        prev_block = NULL;
      }
      next_block = NULL;
    }
    /* decrement the token */
    cur_token = cur_token->prev;
    token = cur_token->data;
    pos -= token->length;
    if (cur_block && cur_token->next == block->begin) {
      next_block = cur_block;
      prev_block = cur_block->prev;
      cur_block = cur_block->parent;
      if(prev_block)
        prev = prev_block->data;
      if (cur_block)
        block = cur_block->data;
    }
    if (prev_block && cur_token == prev->end) {
      cur_block = prev_block;
      block = prev;
      if (prev_block->children) {
        prev_block = g_node_last_sibling(prev_block->children);
        prev = prev_block->data;
      } else {
        prev_block = NULL;
      }
      next_block = NULL;
    }
  }

  /* return the match */
  token_match->token = cur_token;
  token_match->pos = pos;
  token_match->cur_block = cur_block;
  token_match->prev_block = prev_block;
  token_match->next_block = next_block;
}

/* merge_token

   This function merges a token into the token list. The merged token is
   returned.
*/
GList *
_gtk_editor_merge_token (int id, char *text, int length, int g_pos)
{
  GNode *parent = NULL;
  GList *temp, *cur = NULL;
  GSList *text_changes = NULL;
  Block *temp_block;
  Token *token;
  unsigned int pos = 0, len, cur_pos = 0;
  int i, block_id, prop_id;

  g_return_val_if_fail (g_token != NULL, NULL);

  /* update current block */
  if (g_cur_block && g_token->prev == g_block->end) {
    g_prev_block = g_cur_block;
    g_next_block = g_cur_block->next;
    g_cur_block = g_cur_block->parent;
    if(g_cur_block)
      g_block = g_cur_block->data;
    if(g_next_block)
      g_next = g_next_block->data;
  }

  /* Check how the token should be highlighted. This includes checking
     whether or not the token is or is in a block.
  */
  if (((block_id = g_editor->token_entries[id].begin) != -1)
      && ((g_cur_block && g_editor->block_entries[g_block->id].nest
           && g_editor->block_entries[block_id].precedence >=
              g_editor->block_entries[g_block->id].precedence)
          || !g_cur_block)) {
    prop_id = g_editor->block_entries[block_id].prop_id;
  } else if (g_cur_block)
    prop_id = g_editor->block_entries[g_block->id].prop_id;
  else
    prop_id = g_editor->token_entries[id].prop_id;

  /* Loop through the tokens that intersect the merged token. */
  while (pos < length  && g_token) {
    token = g_token->data;
    if (length - pos >= token->length) {
      /* new token overlaps the current token */
      len = token->length;
      if(token->prop_id == -1) {
        /* This token represents new text, we must add it's property to the
           global change list. This is later applied by hilite_buffer
           and insert_text.
        */
        g_new_text_changes = _gtk_editor_add_property_change (g_new_text_changes,
							      prop_id,
							      g_pos+pos,
							      len);
      } else if (token->prop_id != prop_id) {
        text_changes = _gtk_editor_add_property_change (text_changes,
							prop_id,
							g_pos+pos,
							len);
      }

      if (cur == NULL) {
        /* Then merged token doesn't belong to the list yet. Since we would be
           getting rid of this token normally, we can let the merged
           token take over it's spot.
        */
        cur = g_token;
	cur_pos = g_pos;
        token->id = id;
        token->prop_id = prop_id;
        token->length = length;
	g_token = g_token->next;
      } else {
        /* The token no longer exist, recycle it. */
        temp = g_token;
	if (g_token->prev)
	  g_token->prev->next = g_token->next;
	if (g_token->next)
	  g_token->next->prev = g_token->prev;
	g_token = g_token->next;
	temp->next = NULL;
	temp->prev = NULL;
        g_free_tokens = g_list_concat (temp, g_free_tokens);
      }

    } else {
      /* The merged token covers less than the current token. */
      len = length - pos;
      if(token->prop_id == -1) {
        g_new_text_changes = _gtk_editor_add_property_change (g_new_text_changes,
							      prop_id,
							      g_pos+pos,
							      len);
      } else if (token->prop_id != prop_id) {
        text_changes = _gtk_editor_add_property_change (text_changes,
							prop_id,
							g_pos+pos,
							len);
      }
			
      if (cur == NULL) {
        /* Then merged token doesn't belong to the list yet. Since the merged
           token covers less tha this token we have to split it.
        */
	cur = g_token;
	cur_pos = g_pos;
	split_token(g_token, len);
	token->id = id;
	token->prop_id = prop_id;
	token->length = length;
	g_token = g_token->next;
	
      } else {
	/* shorten token */
	token->length -= len;
      }
    }
    /* increment the ammount merged */
    pos += len;
  }

  if (g_cur_block) {
    if (g_editor->token_entries[id].ends) {
      for(parent = g_cur_block; parent != NULL; parent = parent->parent) {
        temp_block = parent->data;
        for (i = 0; i < g_editor->token_entries[id].n_ends; i++) {
          if (temp_block->id == g_editor->token_entries[id].ends[i]) {
            temp_block->end = cur;
            if (!g_block_token) {
              g_base_block = g_cur_block;
              g_base_prev = g_prev_block;
              g_base_next = g_next_block;
              g_block_token = cur;
              g_block_token_pos = g_pos;
            }
            if (g_next_block && cur == g_next->begin) {
              g_cur_block = g_next_block;
              g_block = g_next;
              g_prev_block = NULL;
              g_next_block = g_next_block->children;
              if (g_next_block)
                g_next = g_next_block->data;
            }
            goto out;
          }
        }
      }
    }
  } else if (((block_id = g_editor->token_entries[id].begin) != -1)
             && ((g_cur_block && g_editor->block_entries[g_block->id].nest
                  && g_editor->block_entries[block_id].precedence >=
                     g_editor->block_entries[g_block->id].precedence)
                 || !g_cur_block)) {
    if (g_next_block && cur == g_next->begin) {
      g_cur_block = g_next_block;
      g_block = g_next;
      g_prev_block = NULL;
      g_next_block = g_next_block->children;
      if(g_next_block)
        g_next = g_next_block->data;
      goto out;
    }
    if (g_cur_block)
      parent = g_cur_block->parent;
    else
      parent = NULL;
    g_cur_block = _gtk_editor_get_new_block ();
    if (parent && parent->children == NULL)
      parent->children = g_cur_block;
    if (!g_editor->blocks)
      g_editor->blocks = g_cur_block;
    if (g_prev_block)
      g_prev_block->next = g_cur_block;
    if (g_next_block)
      g_next_block->prev = g_cur_block;
    g_cur_block->next = g_next_block;
    g_cur_block->prev = g_prev_block;
    g_cur_block->parent = parent;
    g_cur_block->children = NULL;
    g_block = g_cur_block->data;
    g_block->begin = cur;
    g_block->end = NULL;
    g_block->id = block_id;
    if (g_next_block)
      g_next = g_next_block->data;
    if (!g_block_token) {
      g_base_block = g_cur_block;
      g_base_prev = g_prev_block;
      g_base_next = g_next_block;
      g_block_token = cur->next;
      g_block_token_pos = g_pos+length;
    }
  }
  out:

  /* apply property changes */
  if(text_changes) {
    _gtk_editor_apply_property_changes(g_editor, text, text_changes, g_pos);
  }

  return cur;
}

static void
split_token (GList *token, int pos)
{
  GList *new_token;
  Token *a, *b;
  
  new_token = get_new_token();
  new_token->next = token->next;
  new_token->prev = token;
  if (token->next)
    token->next->prev = new_token;
  token->next = new_token;
  a = token->data;
  b = new_token->data;
  b->length = a->length - pos;
  b->id = a->id;
  b->prop_id = a->prop_id;
  a->length = pos;
}

static GList*
get_new_token (void)
{
  static GList *temp;
  Token *token;
  
  if (g_free_tokens) {
    temp = g_free_tokens;
    g_free_tokens = g_list_remove_link (g_free_tokens, g_free_tokens);
  } else {
    token = g_new (Token, 1);
    temp = g_new (GList, 1);
    temp->data = token;
  }
  temp->next = NULL;
  temp->prev = NULL;
  return temp;
}

GNode *
_gtk_editor_get_new_block (void)
{
  static GNode *temp;
  Block *block;

  if (g_free_blocks) {
    temp = g_free_blocks;
    g_free_blocks = temp->next;
    temp->next = NULL;
  } else {
    block = g_new (Block, 1);
    temp = g_new (GNode, 1);
    temp->data = block;
  }
  temp->next = NULL;
  temp->prev = NULL;
  return temp;
}

/*
  for debugging
*/
#if DEBUG
static void
print_tokens (GtkEditor *editor)
{
  GList *cur_token;
  GNode *cur_block = editor->blocks;
  Block *block;
  Token *token;
  int i, pos = 0, level = 0;

  if(cur_block)
    block = cur_block->data;
  else
    block = NULL;

  g_print("\n");
  for (cur_token = editor->tokens; cur_token != NULL; cur_token = cur_token->next) {
    token = cur_token->data;
    g_print("|");
    if(cur_token == editor->cur_token)
      g_print("*->|");
    if(block && cur_token == block->begin) {
      if(cur_block == editor->cur_block)
        g_print("C");
      else if(cur_block == editor->prev_block)
        g_print("P");
      else if(cur_block == editor->next_block)
        g_print("N");
      g_print("%d>|", level++);
    }
    for(i = 0; i < token->length; i++) {
      g_print("%c", GTK_TEXT_INDEX(GTK_TEXT(editor), pos + i));
    }
    if(block && cur_token == block->end) {
      cur_block = cur_block->next;
      if(cur_block)
        block = cur_block->data;
      else
        block = NULL;
      g_print("|<%d", --level);
    }
    pos += token->length;
  }
  g_print("|\n");
}

#endif
